package org.bukkit.command;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import org.apache.commons.lang.Validate;
import org.bukkit.Server;
import org.bukkit.command.defaults.AchievementCommand;
import org.bukkit.command.defaults.ClearCommand;
import org.bukkit.command.defaults.SaveOffCommand;
import org.bukkit.command.defaults.TimeCommand;
import org.bukkit.util.Java15Compat;

public class SimpleCommandMap implements CommandMap
{
  private static final Pattern PATTERN_ON_SPACE = Pattern.compile(" ", 16);
  protected final Map<String, Command> knownCommands = new java.util.HashMap();
  private final Server server;
  
  public SimpleCommandMap(Server server) {
    this.server = server;
    setDefaultCommands();
  }
  
  private void setDefaultCommands() {
    register("bukkit", new org.bukkit.command.defaults.SaveCommand());
    register("bukkit", new org.bukkit.command.defaults.SaveOnCommand());
    register("bukkit", new SaveOffCommand());
    register("bukkit", new org.bukkit.command.defaults.StopCommand());
    register("bukkit", new org.bukkit.command.defaults.VersionCommand("version"));
    register("bukkit", new org.bukkit.command.defaults.ReloadCommand("reload"));
    register("bukkit", new org.bukkit.command.defaults.PluginsCommand("plugins"));
    register("bukkit", new org.bukkit.command.defaults.TimingsCommand("timings"));
  }
  
  public void setFallbackCommands() {
    register("bukkit", new org.bukkit.command.defaults.ListCommand());
    register("bukkit", new org.bukkit.command.defaults.OpCommand());
    register("bukkit", new org.bukkit.command.defaults.DeopCommand());
    register("bukkit", new org.bukkit.command.defaults.BanIpCommand());
    register("bukkit", new org.bukkit.command.defaults.PardonIpCommand());
    register("bukkit", new org.bukkit.command.defaults.BanCommand());
    register("bukkit", new org.bukkit.command.defaults.PardonCommand());
    register("bukkit", new org.bukkit.command.defaults.KickCommand());
    register("bukkit", new org.bukkit.command.defaults.TeleportCommand());
    register("bukkit", new org.bukkit.command.defaults.GiveCommand());
    register("bukkit", new TimeCommand());
    register("bukkit", new org.bukkit.command.defaults.SayCommand());
    register("bukkit", new org.bukkit.command.defaults.WhitelistCommand());
    register("bukkit", new org.bukkit.command.defaults.TellCommand());
    register("bukkit", new org.bukkit.command.defaults.MeCommand());
    register("bukkit", new org.bukkit.command.defaults.KillCommand());
    register("bukkit", new org.bukkit.command.defaults.GameModeCommand());
    register("bukkit", new org.bukkit.command.defaults.HelpCommand());
    register("bukkit", new org.bukkit.command.defaults.ExpCommand());
    register("bukkit", new org.bukkit.command.defaults.ToggleDownfallCommand());
    register("bukkit", new org.bukkit.command.defaults.BanListCommand());
    register("bukkit", new org.bukkit.command.defaults.DefaultGameModeCommand());
    register("bukkit", new org.bukkit.command.defaults.SeedCommand());
    register("bukkit", new org.bukkit.command.defaults.DifficultyCommand());
    register("bukkit", new org.bukkit.command.defaults.WeatherCommand());
    register("bukkit", new org.bukkit.command.defaults.SpawnpointCommand());
    register("bukkit", new ClearCommand());
    register("bukkit", new org.bukkit.command.defaults.GameRuleCommand());
    register("bukkit", new org.bukkit.command.defaults.EnchantCommand());
    register("bukkit", new org.bukkit.command.defaults.TestForCommand());
    register("bukkit", new org.bukkit.command.defaults.EffectCommand());
    register("bukkit", new org.bukkit.command.defaults.ScoreboardCommand());
    register("bukkit", new org.bukkit.command.defaults.PlaySoundCommand());
    register("bukkit", new org.bukkit.command.defaults.SpreadPlayersCommand());
    register("bukkit", new org.bukkit.command.defaults.SetWorldSpawnCommand());
    register("bukkit", new org.bukkit.command.defaults.SetIdleTimeoutCommand());
    register("bukkit", new AchievementCommand());
  }
  


  public void registerAll(String fallbackPrefix, List<Command> commands)
  {
    if (commands != null) {
      for (Command c : commands) {
        register(fallbackPrefix, c);
      }
    }
  }
  


  public boolean register(String fallbackPrefix, Command command)
  {
    return register(command.getName(), fallbackPrefix, command);
  }
  


  public boolean register(String label, String fallbackPrefix, Command command)
  {
    label = label.toLowerCase().trim();
    fallbackPrefix = fallbackPrefix.toLowerCase().trim();
    boolean registered = register(label, command, false);
    knownCommands.put(fallbackPrefix + ":" + label, command);
    
    Iterator<String> iterator = command.getAliases().iterator();
    while (iterator.hasNext()) {
      if (!register((String)iterator.next(), command, true)) {
        iterator.remove();
      }
    }
    

    if (!registered) {
      command.setLabel(fallbackPrefix + ":" + label);
    }
    

    command.register(this);
    
    return registered;
  }
  






  private synchronized boolean register(String label, Command command, boolean isAlias)
  {
    if ((isAlias) && (knownCommands.containsKey(label)))
    {

      return false;
    }
    
    boolean registered = true;
    

    Command conflict = (Command)knownCommands.get(label);
    if ((conflict != null) && (conflict.getLabel().equals(label))) {
      return false;
    }
    
    if (!isAlias) {
      command.setLabel(label);
    }
    knownCommands.put(label, command);
    
    return registered;
  }
  

  public boolean dispatch(CommandSender sender, String commandLine)
    throws CommandException
  {
    String[] args = PATTERN_ON_SPACE.split(commandLine);
    
    if (args.length == 0) {
      return false;
    }
    
    String sentCommandLabel = args[0].toLowerCase();
    Command target = getCommand(sentCommandLabel);
    
    if (target == null) {
      return false;
    }
    
    try
    {
      target.execute(sender, sentCommandLabel, (String[])Java15Compat.Arrays_copyOfRange(args, 1, args.length));
    } catch (CommandException ex) {
      throw ex;
    } catch (Throwable ex) {
      throw new CommandException("Unhandled exception executing '" + commandLine + "' in " + target, ex);
    }
    

    return true;
  }
  
  public synchronized void clearCommands() {
    for (Map.Entry<String, Command> entry : knownCommands.entrySet()) {
      ((Command)entry.getValue()).unregister(this);
    }
    knownCommands.clear();
    setDefaultCommands();
  }
  
  public Command getCommand(String name) {
    Command target = (Command)knownCommands.get(name.toLowerCase());
    return target;
  }
  
  public List<String> tabComplete(CommandSender sender, String cmdLine) {
    Validate.notNull(sender, "Sender cannot be null");
    Validate.notNull(cmdLine, "Command line cannot null");
    
    int spaceIndex = cmdLine.indexOf(' ');
    
    if (spaceIndex == -1) {
      ArrayList<String> completions = new ArrayList();
      Map<String, Command> knownCommands = this.knownCommands;
      
      String prefix = (sender instanceof org.bukkit.entity.Player) ? "/" : "";
      
      for (Map.Entry<String, Command> commandEntry : knownCommands.entrySet()) {
        Command command = (Command)commandEntry.getValue();
        
        if (command.testPermissionSilent(sender))
        {


          String name = (String)commandEntry.getKey();
          
          if (org.bukkit.util.StringUtil.startsWithIgnoreCase(name, cmdLine)) {
            completions.add(prefix + name);
          }
        }
      }
      Collections.sort(completions, String.CASE_INSENSITIVE_ORDER);
      return completions;
    }
    
    String commandName = cmdLine.substring(0, spaceIndex);
    Command target = getCommand(commandName);
    
    if (target == null) {
      return null;
    }
    
    if (!target.testPermissionSilent(sender)) {
      return null;
    }
    
    String argLine = cmdLine.substring(spaceIndex + 1, cmdLine.length());
    String[] args = PATTERN_ON_SPACE.split(argLine, -1);
    try
    {
      return target.tabComplete(sender, commandName, args);
    } catch (CommandException ex) {
      throw ex;
    } catch (Throwable ex) {
      throw new CommandException("Unhandled exception executing tab-completer for '" + cmdLine + "' in " + target, ex);
    }
  }
  
  public java.util.Collection<Command> getCommands() {
    return Collections.unmodifiableCollection(knownCommands.values());
  }
  
  public void registerServerAliases() {
    Map<String, String[]> values = server.getCommandAliases();
    
    for (String alias : values.keySet()) {
      if ((alias.contains(":")) || (alias.contains(" "))) {
        server.getLogger().warning("Could not register alias " + alias + " because it contains illegal characters");
      }
      else
      {
        String[] commandStrings = (String[])values.get(alias);
        List<String> targets = new ArrayList();
        StringBuilder bad = new StringBuilder();
        
        for (String commandString : commandStrings) {
          String[] commandArgs = commandString.split(" ");
          Command command = getCommand(commandArgs[0]);
          
          if (command == null) {
            if (bad.length() > 0) {
              bad.append(", ");
            }
            bad.append(commandString);
          } else {
            targets.add(commandString);
          }
        }
        
        if (bad.length() > 0) {
          server.getLogger().warning("Could not register alias " + alias + " because it contains commands that do not exist: " + bad);



        }
        else if (targets.size() > 0) {
          knownCommands.put(alias.toLowerCase(), new FormattedCommandAlias(alias.toLowerCase(), (String[])targets.toArray(new String[targets.size()])));
        } else {
          knownCommands.remove(alias.toLowerCase());
        }
      }
    }
  }
}
